/*
 * $Id: forth.S,v 1.171 2000/08/04 19:53:25 andrewm Exp $
 *
 * PURPOSE
 *	Embedable Forth for little endian StrongARM (armv4l).
 *
 * DESCRIPTION
 *	This Forth was inspired by eForth, but is much more closely related to
 *	ANS Forth.  My goal is to create a fast, full featured StrongARM Forth
 *	that can be used in a firmware through a serial console.  As a result,
 *	this Forth uses indirectly threaded code, unlike eForth's directly
 *	threaded code.  The word set is also _much_ richer than eForth's
 *	minimalistic approach.
 *
 *	This has been, and likely will continue to be, a learning experience
 *	for me.  I knew little about Forth starting out, and have learned a
 *	whole lot as a result.  My ARM assembly programming ability has also
 *	improved quite a bit since my days of simple inline C assembly.
 *
 *	In the course of writing this I discovered bugs in the GNU binutils,
 *	which the GNU "as" (a.k.a. "gas") assembler is a part of:
 *	   1) .byte directives were assembled as .long
 *	   2) .macro string arguments could not contain escaped characters
 *	   3) The objdump utility used a different register naming convention
 *	      than GNU gcc.
 *	All of the above bugs have been fixed in the official GNU binutils
 *	distribution as a result of me discovering them.
 *
 *	All development was done on a StrongARM based NetWinder DM available
 *	from Rebel.com (http://www.rebel.com/).  This Forth was intended to be
 *	used in the firmware of this Linux machine.  Having a full featured
 *	native development environment made the task much easier.
 *
 *	Special thanks to Forth gurus Ralph Siemsen and Eric LaForest for their
 *	advice and encouragement, and to ARM tools guru Scott Bambrough for
 *	putting up with my obscure ARM questions.
 *
 *	- aligned addresses are a multiple of 4 address units (bytes).
 *	- EMIT outputs all 8-bit characters as-is.
 *	- character addresses are a multiple of 1 address unit (byte).
 *	- conversion of digits larger than 35 ("Z") is unpredictable.
 *	- maximum size of a counted string is 255 characters.
 *	- the variable `KEY is a pointer to the input xt.
 *	- the variable `EMIT is a pointer to the output xt.
 *	- addresses are 32 bits.
 *	- signed numbers are in twos complement format.
 *	- numeric ranges by type:
 *		-(2^31) <=	n	<= (2^31 - 1)
 *		0	<=	n+	<= (2^31 - 1)
 *		0	<=	u	<= (2^32 - 1)
 *		-(2^63) <=	d	<= (2^63 - 1)
 *		0	<=	d+	<= (2^63 - 1)
 *		0	<=	ud	<= (2^64 - 1)
 *	- addresses are 4 address units (bytes)
 *	- characters are 1 address unit (byte)
 *
 *	If you modify the code or its order, make sure to run the perl linked
 *	list updating script link.pl, or manually adjust the links yourself.
 *	[Note: the GNU assembler (GAS), forces me to use this method.]
 *
 * HISTORY
 *	August 4, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	DOES> is useable now.  At this time there are 276 words.
 *	Public alpha release (unannounced): saforth-0.12
 *
 *	August 3, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Public alpha release (announced): saforth-0.11
 *
 *	August 1, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Numerous changes to case and loop structures.
 *	Public alpha release (unannounced): saforth-0.10
 *
 *	July 31, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	First public alpha release (announced): saforth-0.9
 *
 *	July 28, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Colon definitions work, though DOES> cannot be handled yet.
 *
 *	July 17, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Can do some rudimentary interpretation.  At this time there are 248
 *	words.
 *
 *	July 10, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Finished WORD, though it could stand some cleaning up.  Let the parsing
 *	begin!
 *
 *	July  4, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Initial conversion of VARIABLE to use offsets for UP, with register
 *	HERE being assigned UP's previous job of pointing to the next free
 *	address in the data space.  Removed CONSTANT macro.  Rewrote words as
 *	colon definitions, when possible, that directly referenced variables or
 *	constants in primitives.
 *
 *	June 28, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Finished all control flow related words.  At this time there are 190
 *	words.
 *
 *	June 23, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Tested dictionary traversal (used link.pl to generate links).  At this
 *	time there are 167 words, though some will have to be looked at again
 *	since they don't take the compile state into account.
 *
 *	June 14, 2000 - Andrew E. Mileski <andrewm@netwinder.org>
 *	Finished threading (ie. have a working Forth machine) and began testing.
 *	At this time there are 78 words.  More words will be added as a compiler
 *	and interpreter are built.  Began documenting in more detail.
 *
 *	Copyright 2000, Andrew E. Mileski <andrewm@netwinder.org>
 */

/* Words to include */
#define CORE_EXT
#undef	BLOCK
#undef	BLOCK_EXT
#define DOUBLE
#define DOUBLE_EXT
#undef	FACILITY
#undef	FACILITY_EXT
#undef	FILE
#undef	FILE_EXT
#define	FIXED
#undef	LOCAL
#undef	LOCAL_EXT
#undef	MEMORY
#define	TOOLS
#define	TOOLS_EXT
#undef	SEARCH
#undef	SEARCH_EXT
#define	STRING
#define FIXED_EXT

/* Memory are sizes (not all are currently used */
#define CP_SIZE		(128 * 4)
#define	HOLD_SIZE	128
#define	PAD_SIZE	256
#define SP_SIZE		(128 * 4)
#define RP_SIZE		(128 * 4)
#define	TIB_SIZE	256
#define WORD_SIZE	256

/* Working registers - the order must be ascending */
#define	TOS		r0	/* Top of parameter stack (SP) */
#define TMP0		r1	/* Temporary (doubles as W) */
#define TMP1		r2	/* Temporary */
#define TMP2		r3	/* Temporary */
#define TMP3		r4	/* Temporary */
#define TMP4		r5	/* Temporary */
#define TMP5		r6	/* Temporary */

/* Miscellaneous registers - the order doesn't matter */
#define	UP		r7	/* User area pointer */
#define HERE		r8	/* Dictionary pointer */
#define RP		r9	/* Return stack pointer */
#define CP		r10	/* Control flow stack pointer */
				/* fp (r11) is reserved */
#define IP		ip	/* Current instruction pointer */
#define SP		sp	/* Parameter stack pointer - last item */
				/* lr (r14) is unused */
				/* pc (r15) is reserved */

#define NOT_SAVED	r0, r1, r2, r3, ip
#define SAVED		r4, r5, r6, r7, r8, r9, sl /* fp, sp, lr */

/* Boolean */
#define	TRUE		-1
#define	FALSE		0

/* States */
#define S_INTERPRET	0
#define	S_COMPILE	1
#define S_DOES		2

/* Flags - note that link.pl parses using spaces so none are allowed here */
#define	COMPILE_ONLY	0x01	/* Not available when STATE = 0 */
#define	IMMEDIATE	0x02	/* Execute when STATE <> 0 */
#define	IMMEDIATE_ONLY	(COMPILE_ONLY|IMMEDIATE)

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Macro Definitions
 *
 *	ALIAS		Creates an alias.
 *	CODE_BEGIN	Mark the start of a code fragment.
 *	CODE_END	Mark the end of a code fragment.
 *	COLON_BEGIN	Mark the start of a colon definition.
 *	COLON_END	Mark the end of a colon definition.
 *	CSTRING		Define a counted string
 *	DICT_BEGIN	Mark the start of a dictionary definition.
 *	DICT_END	Mark the end of a dictionary definition.
 *	NEXT		Execute the next Forth instruction (IP += 4).
 *	VARIABLE	Creates a variable.
 */
	.macro	ALIAS		flags, name, label, alias, link
		.data
		.align	2
_first\label:	.long	_first\link
		.byte	\flags
		.byte	(2f - 1f)
1:		.ascii	"\name"
2:		.align	2
\label:		.long	_code\alias
	.endm

	.macro	CODE_BEGIN	lex, name, label, link
		ALIAS		\lex, "\name", \label, \label, \link
		.text
		.code	32
		.align	2
_code\label:
	.endm

	.macro	CODE_END
	.endm

	.macro	COLON_BEGIN	lex, name, label, link
		ALIAS		\lex, "\name", \label, _enter, \link
		.data
		.align	2
_thread\label:
	.endm

	.macro	COLON_END
	.endm

	.macro	DICT_BEGIN
		.set	_firstFIRST, 0
		.data
		.align	2
		.text
		.align	5
		.code	32
	.endm

	.macro	DICT_END
		.data
		.set	EDATA, .
		.text
		.set	ETEXT, .
	.endm

	.macro	CSTRING		str, label
		.data
		.ifc	"L", "L\label"
			.byte		(2f - 1f)
		.else
_cstr\label:		.byte		(2f - 1f)
		.endif
1:		.ascii		"\str"
2:
	.endm

	.macro	NEXT	cc="al", preinc=4
		ldr\cc	TMP0, [IP, #\preinc]!
		ldr\cc	pc, [TMP0]
	.endm

	.macro	VARIABLE	lex, name, label, link
		ALIAS	\lex, "\name", \label, _dovar, \link
		.data
		.align	2
_var\label:	.long	_off\label
	.endm

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Startup
 */

	.global	forth

	/* Setup a stack frame */
forth:	mov	ip, sp
	stmfd	sp!, {r4, r5, r6, r7, r8, r9, sl, fp, ip, lr}
	mov	fp, sp

	/* Initialize UP */
	add	r0, r0, #3
	bic	UP, r0, #3
	str	UP, [UP, #_off_up0]

	/* Initialize UPTOP */
	add	r1, UP, r1
	str	r1, [UP, #_off_uptop]

	/* Copy initial values */
cold:	mov	HERE, UP
	ldr	r4, =user_start
	ldr	r5, =user_end
1:	ldr	r6, [r4], #4
	str	r6, [HERE], #4
	cmp	r4, r5
	blo	1b

	/* Initialize SP */
	add	HERE, HERE, #SP_SIZE
	mov	SP, HERE
	str	SP, [UP, #_off_sp0]
	ldr	TOS, =0xdeadbeef

	/* Initialize RP */
	add	HERE, HERE, #RP_SIZE
	mov	RP, HERE
	str	RP, [UP, #_off_rp0]
	
	/* Initialize ABORT as the first token */
	ldr	TMP0, =_abort

	/* Execute ABORT - shouldn't return as it nukes the return stack */
	ldr	pc, [TMP0]
	/* End */

gbye:	mov	ip, fp
	ldmfd	ip, {r4, r5, r6, r7, r8, r9, sl, fp, sp, pc}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Division routines
 */

	/*
	 * Subroutine for unsigned symmetric 32 bit division
	 *
	 * TOS  = divisor
	 * TMP0 = dividend and quotient
	 * TMP1 = remainder
	 * TMP2 = bit count
	 */
udivide32:	mov	TMP1, #0
		mov	TMP2, #32
1:		movs	TMP0, TMP0, lsl #1
		adc	TMP1, TMP1, TMP1
		cmp	TMP1, TOS
		subhs	TMP1, TMP1, TOS
		orrhs	TMP0, TMP0, #1
		subs	TMP2, TMP2, #1
		bne	1b
	/* Done */
		mov	pc, lr
	/* End */

	/*
	 * Subroutine for signed symmetric 32 bit division
	 *
	 * TOS  = divisor
	 * TMP0 = dividend and quotient
	 * TMP1 = remainder
	 * TMP2 = bit count
	 * TMP3 = sign flags
	 * TMP4 = return address
	 */
divide32:	mov	TMP4, lr	
		mov	TMP3, #0
	/* Make divisor unsigned (positive) */
		tst	TOS, #(1 << 31)
		rsbne	TOS, TOS, #0
		eorne	TMP3, TMP3, #1
	/* Make dividend unsigned (positive) */
		tst	TMP0, #(1 << 31)
		rsbne	TMP0, TMP0, #0
		eorne	TMP3, TMP3, #3
	/* Divide */
		bl	udivide32
	/* Sign the quotient */
		tst	TMP3, #1
		rsbne	TMP0, TMP0, #0
	/* Sign the remainder */
		tst	TMP3, #2
		rsbne	TMP1, TMP1, #0
		mov	pc, TMP4
	/* Done */

	/*
	 * Subroutine for unsigned symmetric 64 bit division
	 *
	 * TOS  = divisor high
	 * TMP0 = divisor low
	 * TMP1 = dividend high
	 * TMP2 = dividend low
	 * TMP3 = remainder high
	 * TMP4 = remainder low
	 * TMP5 = count
	 */
udivide64:	mov	TMP3, #0
		mov	TMP4, #0
		mov	TMP5, #64
1:		movs	TMP2, TMP2, lsl #1
		adcs	TMP1, TMP1, TMP1
		adcs	TMP4, TMP4, TMP4
		adc	TMP3, TMP3, TMP3
		cmp	TMP3, TOS
		cmpeq	TMP4, TMP0 
		blo	2f
		subs	TMP4, TMP4, TMP0
		sbc	TMP3, TMP3, TOS
		orr	TMP2, TMP2, #1
2:		subs	TMP5, TMP5, #1
		bne	1b
	/* Done */
		mov	pc, lr
	/* End */

	/*
	 * Subroutine for signed symmetric 64 bit division
	 *
	 * TOS  = divisor high
	 * TMP0 = divisor low
	 * TMP1 = dividend high and quotient high
	 * TMP2 = dividend low and quotient low
	 * TMP3 = remainder high
	 * TMP4 = remainder low
	 * TMP5 = count
	 * lr	= return address and sign flags
	 */
divide64:	str	lr, [SP, #-4]!
		mov	lr, #0
	/* Make divisor unsigned (positive) */
		tst	TOS, #(1 << 31)
		beq	1f
		rsbs	TMP0, TMP0, #0
		rsc	TOS, TOS, #0
		eor	lr, lr, #1
	/* Make dividend unsigned (positive) */
1:		tst	TMP1, #(1 << 31)
		beq	2f
		rsbs	TMP2, TMP2, #0
		rsc	TMP1, TMP1, #0
		eor	lr, lr, #3
	/* Divide */
2:		bl	udivide64
	/* Sign the quotient */
		tst	lr, #1
		beq	3f
		rsbs	TMP2, TMP2, #0
		rsc	TMP1, TMP1, #0
	/* Sign the remainder */
3:		tst	lr, #2
		beq	4f
		rsbs	TMP4, TMP4, #0
		rsc	TMP3, TMP3, #0
	/* Done */
4:		ldr	pc, [SP], #4
	/* End */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Predefined user variables
 */

	.macro		USER	name, label, value
		.set	_off\label, (. - user_start)
		.long	\value
	.endm

	.data
	.align	2

user_start:
	/* Reserved */
	.rept	32
		.long	0
	.endr

	/* User variables */
	USER	"`ACCEPT", _tick_accept, _doACCEPT
	USER	"`CR", _tick_cr, _doCR
	USER	"`EMIT", _tick_emit, _doEMIT
	USER	"`KEY", _tick_key, _doKEY
	USER	"`PROMPT", _tick_prompt, _doPROMPT
	USER	">IN", _from_in, 0
	USER	"BASE", _base, 10
	USER	"CP0", _cp0, 0
	USER	"ERR", _err, 0
	USER	"FOUND", _found, 0
	USER	"HOLD0", _hold0, 0
	USER	"HOLDP", _holdp, 0
	USER	"LAST", _last, _first_colon
	USER	"PADP", _padp, 0
	USER	"SP0", _sp0, 0
	USER	"RP0", _rp0, 0
	USER	"STATE", _state, 0
	USER	"#TIB", _number_tib, 0
	USER	"TIB", _tib, 0
user_end:
	USER	"UP0", _up0, 0
	USER	"UPTOP", _uptop, 0

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	COLON_BEGIN	COMPILE_ONLY, "doACCEPT", _doACCEPT, _doACCEPT
		.long	_dup, _to_r
		.long	_swap
1:		.long	_to_r
		.long	_dup
		.long	_question_branch, 3f
			.long	_doKEY
			.long	_dup, _lit, '\n', _equal
			.long	_question_branch, 2f
				.long	_drop
				.long	_branch, 3f
2:			.long	_r_fetch, _c_store
			.long	_1_minus
			.long	_r_from, _char_plus
			.long	_branch, 1b
3:		.long	_r_from, _drop
		.long	_r_from, _swap, _minus
		.long	_exit
	COLON_END

	/* Generate new line on default output device */
	COLON_BEGIN	COMPILE_ONLY, "doCR", _doCR, _doCR
		.long	_lit, '\n'
		.long	_doEMIT
		.long	_exit
	COLON_END

	/* Default output device */
	CODE_BEGIN	COMPILE_ONLY, "doEMIT", _doEMIT, _doEMIT
		stmfd	sp!, {NOT_SAVED}
		mov	r0, #1		/* stdout */
		mov	r1, sp		/* address of TOS */
		mov	r2, #1		/* 1 character */
		swi	0x00900004	/* sys_write */
		ldmfd	sp!, {NOT_SAVED}
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	/* Default input device */
	CODE_BEGIN	COMPILE_ONLY, "doKEY", _doKEY, _doKEY
		str	TOS, [SP, #-4]!
		mov	TOS, #0
		stmfd	sp!, {NOT_SAVED}
		mov	r0, #0		/* stdin */
		mov	r1, sp		/* address of TOS */
		mov	r2, #1		/* 1 character */
		swi	0x00900003	/* sys_read */
		ldmfd	sp!, {NOT_SAVED}
		NEXT
	CODE_END

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

DICT_BEGIN

/*
 * Implementation specific
 */
	CODE_BEGIN	0, ">C", _to_c, FIRST
		str	TOS, [CP, #-4]!
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "C>", _c_from, _to_c
		str	TOS, [SP, #-4]!
		ldr	TOS, [CP], #4
		NEXT
	CODE_END

	/* ( * -- ) Re-initialize everything */
	CODE_BEGIN	0, "COLD", _cold, _c_from
		b	cold
	CODE_END

	/* ( -- a-addr ) Address of cell containing address of initial CP */
	VARIABLE	0, "CP0", _cp0, _cold

	/* ( -- a-addr ) Push the control flow stack pointer */
	CODE_BEGIN	0, "CP@", _cp_fetch, _cp0
		str	TOS, [SP, #-4]!
		mov	TOS, CP
		NEXT
	CODE_END

	/* ( a-addr -- ) Load a new control flow stack pointer */
	CODE_BEGIN	0, "CP!", _cp_store, _cp_fetch
		mov	CP, TOS
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	/* ( R: -- a-addr ) Enter into a thread */
	CODE_BEGIN	COMPILE_ONLY, "ENTER", _enter, _cp_store
		str	IP, [RP, #-4]!
		mov	IP, TMP0	/* IP + 4 = new thread */
		NEXT
	CODE_END

	/* ( -- a-addr ) Address of cell containing address of last word */
	VARIABLE	0, "LAST", _last, _enter

	/* ( -- ) No-operation */
	CODE_BEGIN	0, "NOP", _nop, _last
		NEXT
	CODE_END

	/* ( a-addr -- ) Load a new parameter stack pointer */
	CODE_BEGIN	0, "SP!", _sp_store, _nop
		mov	SP, TOS
		NEXT
	CODE_END

	/* ( -- a-addr ) Address of cell containing address of initial SP */
	VARIABLE	0, "SP0", _sp0, _sp_store

	/* ( -- a-addr ) Push the parameter stack pointer */
	CODE_BEGIN	0, "SP@", _sp_fetch, _sp0
		str	TOS, [SP, #-4]!
		mov	TOS, SP
		NEXT
	CODE_END

	/* ( a-addr -- ) Load a new return stack pointer */
	CODE_BEGIN	0, "RP!", _rp_store, _sp_fetch
		mov	RP, TOS
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	/* ( -- a-addr ) Address of cell containing address of initial RP */
	VARIABLE	0, "RP0", _rp0, _rp_store

	/* ( -- a-addr ) Push the return stack pointer */
	CODE_BEGIN	0, "RP@", _rp_fetch, _rp0
		str	TOS, [SP, #-4]!
		mov	TOS, RP
		NEXT
	CODE_END

	CODE_BEGIN	COMPILE_ONLY, "S<>C", _s_swap_c, _rp_fetch
		str	TOS, [SP, #-4]!
		eor	SP, SP, CP
		eor	CP, SP, CP
		eor	SP, SP, CP
		ldr	TOS, [SP], #4
		ldr	TMP0, [UP, #_off_sp0]
		ldr	TMP1, [UP, #_off_cp0]
		str	TMP1, [UP, #_off_sp0]
		str	TMP0, [UP, #_off_cp0]
		NEXT
	CODE_END

	CODE_BEGIN	COMPILE_ONLY, "S<>R", _s_swap_r, _s_swap_c
		str	TOS, [SP, #-4]!
		eor	SP, SP, RP
		eor	RP, SP, RP
		eor	SP, SP, RP
		ldr	TOS, [SP], #4
		ldr	TMP0, [UP, #_off_sp0]
		ldr	TMP1, [UP, #_off_rp0]
		str	TMP1, [UP, #_off_sp0]
		str	TMP0, [UP, #_off_rp0]
		NEXT
	CODE_END

	CODE_BEGIN	COMPILE_ONLY, "SKIP", _skip, _s_swap_r
		NEXT	al, 8
	CODE_END

	/* ( -- a-addr ) Address of cell containing number of chars in  TIB */
	VARIABLE	0, "#TIB", _number_tib, _skip

	/* ( -- a-addr ) Address of cell containing address of TIB */
	VARIABLE	0, "TIB", _tib, _number_tib

	/* ( -- a-addr ) Address of cell containing address of UP */
	VARIABLE	0, "UP0", _up0, _tib

	/* ( -- a-addr ) Address of cell containing address of UPTOP */
	VARIABLE	0, "UPTOP", _uptop, _up0

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	CODE_BEGIN	0, "!", _store, _uptop
		ldmfd	SP!, {TMP0, TMP1}
		str	TMP0, [TOS]
		mov	TOS, TMP1
		NEXT
	CODE_END

	CODE_BEGIN	0, "*", _star, _store
		ldr	TMP0, [SP], #4
		mul	TOS, TMP0, TOS
		NEXT
	CODE_END

	CODE_BEGIN	0, "*/", _star_slash, _star
		ldmfd	SP!, {TMP3, TMP4}
		/* Multiply */
		smull	TMP2, TMP1, TOS, TMP4
		/* Sign extend divisor */
		movs	TMP1, TMP3
		movpl	TOS, #0
		mvnmi	TOS, #0
		/* Divide */
		bl	divide64
		/* Only need quotient low */
		mov	TOS, TMP2
		NEXT
	CODE_END

	CODE_BEGIN	0, "*/MOD", _star_slash_mod, _star_slash
		ldmfd	SP!, {TMP3, TMP4}
		/* Multiply */
		smull	TMP2, TMP1, TOS, TMP4
		/* Sign extend divisor */
		movs	TMP1, TMP3
		movpl	TOS, #0
		mvnmi	TOS, #0
		/* Divide */
		bl	divide64
		/* Quotient low and remainder low */
		mov	TOS, TMP2
		str	TMP4, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "+", _plus, _star_slash_mod
		ldr	TMP0, [SP], #4
		add	TOS, TMP0, TOS
		NEXT
	CODE_END

	CODE_BEGIN	0, "+!", _plus_store, _plus
		ldmfd	SP!, {TMP0, TMP1}
		ldr	TMP2, [TOS]
		add	TMP2, TMP2, TMP0
		str	TMP2, [TOS]
		mov	TOS, TMP1
		NEXT
	CODE_END

	/* Implementation specific - runtime for +LOOP */
	CODE_BEGIN	COMPILE_ONLY, "(+LOOP)", _do_plus_loop, _plus_store
		ldmfd	RP!, {TMP0, TMP1}
		adds	TMP0, TMP0, TOS
		ldr	TOS, [SP], #4
		NEXT	vs, 8
		stmfd	RP!, {TMP0, TMP1}
		ldr	IP, [IP, #4]
		NEXT	al, 0
	CODE_END

	CODE_BEGIN	IMMEDIATE, "+LOOP", _plus_loop, _do_plus_loop
		ldr	TMP0, =_do_plus_loop
		str	TMP0, [HERE], #4
		ldr	TMP0, [CP], #4
		str	TMP0, [HERE], #4
1:		teq	TOS, #0
		ldreq	TOS, [SP], #4
		NEXT	eq
		ldr	TMP0, [SP], #4
		str	HERE, [TMP0]
		subs	TOS, TOS, #1
		b	1b
	CODE_END

	CODE_BEGIN	0, ",", _comma, _plus_loop
		str	TOS, [HERE], #4
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "-", _minus, _comma
		ldr	TMP0, [SP], #4
		sub	TOS, TMP0, TOS
		NEXT
	CODE_END

	CODE_BEGIN	0, "/", _slash, _minus
		ldr	TMP0, [SP], #4
		bl	divide32
		mov	TOS, TMP0
		NEXT
	CODE_END

	CODE_BEGIN	0, "/MOD", _slash_mod, _slash
		ldr	TMP0, [SP], #4
		bl	divide32
		mov	TOS, TMP0
		str	TMP1, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "0<", _0_less, _slash_mod
		mov	TOS, TOS, asr #31
		NEXT
	CODE_END

	CODE_BEGIN	0, "0=", _0_equal, _0_less
		teq	TOS, #0
		mov	TOS, #FALSE
		moveq	TOS, #TRUE
		NEXT
	CODE_END

	CODE_BEGIN	0, "1+", _1_plus, _0_equal
		add	TOS, TOS, #1
		NEXT
	CODE_END

	CODE_BEGIN	0, "1-", _1_minus, _1_plus
		sub	TOS, TOS, #1
		NEXT
	CODE_END

	CODE_BEGIN	0, "2!", _2_store, _1_minus
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		stmfd	TOS, {TMP0, TMP1}
		mov	TOS, TMP2
		NEXT
	CODE_END

	CODE_BEGIN	0, "2*", _2_star, _2_store
		mov	TOS, TOS, lsl #1
		NEXT
	CODE_END

	CODE_BEGIN	0, "2/", _2_slash, _2_star
		mov	TOS, TOS, asr #1
		NEXT
	CODE_END

	CODE_BEGIN	0, "2@", _2_fetch, _2_slash
		ldmfd	TOS, {TMP0, TMP1}
		str	TMP1, [SP, #-4]!
		mov	TOS, TMP0
		NEXT
	CODE_END

	CODE_BEGIN	0, "2DROP", _2drop, _2_fetch
		add	SP, SP, #4
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "2DUP", _2dup, _2drop
		ldr	TMP0, [SP]
		mov	TMP1, TOS
		stmfd	SP!, {TMP0, TMP1}
		NEXT
	CODE_END

	CODE_BEGIN	0, "2OVER", _2over, _2dup
		mov	TMP1, TOS
		add     TMP2, SP, #4
		ldmfd   TMP2, {TOS, TMP0}
		stmfd	SP!, {TMP0, TMP1}
		NEXT
	CODE_END

	CODE_BEGIN	0, "2SWAP", _2swap, _2over
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		stmfd	SP!, {TOS, TMP0}
		str	TMP2, [SP, #-4]!
		mov	TOS, TMP1
		NEXT
	CODE_END

	/* Implementation specific - handy in a 32 bit environment */
	CODE_BEGIN	0, "4*", _4_star, _2swap
		mov	TOS, TOS, lsl #2
		NEXT
	CODE_END

	/* Implementation specific - handy in a 32 bit environment */
	CODE_BEGIN	0, "4/", _4_slash, _4_star
		mov	TOS, TOS, asr #2
		NEXT
	CODE_END

	CODE_BEGIN	0, "<", _less, _4_slash
		ldr	TMP0, [SP], #4
		cmp	TMP0, TOS
		mov	TOS, #FALSE
		movlt	TOS, #TRUE
		NEXT
	CODE_END

	CODE_BEGIN	0, "=", _equal, _less
		ldr	TMP0, [SP], #4
		teq	TOS, TMP0
		mov	TOS, #FALSE
		moveq	TOS, #TRUE
		NEXT
	CODE_END

	CODE_BEGIN	0, ">", _greater, _equal
		ldr	TMP0, [SP], #4
		cmp	TMP0, TOS
		mov	TOS, #FALSE
		movgt	TOS, #TRUE
		NEXT
	CODE_END

	CODE_BEGIN	0, ">BODY", _to_body, _greater
		add	TOS, TOS, #4
		NEXT
	CODE_END

	VARIABLE	0, ">IN", _from_in, _to_body

	CODE_BEGIN	0, ">R", _to_r, _from_in
		str	TOS, [RP, #-4]!
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "?DUP", _question_dup, _to_r
		teq	TOS, #0
		strne	TOS, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "@", _fetch, _question_dup
		ldr	TOS, [TOS]
		NEXT
	CODE_END

	CODE_BEGIN	0, "ABS", _abs, _fetch
		tst	TOS, #(1 << 31)
		rsbne	TOS, TOS, #0
		NEXT
	CODE_END

	CODE_BEGIN	0, "ALIGN", _align, _abs
		add	HERE, HERE, #3
		bic	HERE, HERE, #3
		NEXT
	CODE_END

	CODE_BEGIN	0, "ALIGNED", _aligned, _align
		add	TOS, TOS, #3
		bic	TOS, TOS, #3
		NEXT
	CODE_END

	CODE_BEGIN	0, "ALLOT", _allot, _aligned
		add	HERE, HERE, TOS
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "AND", _and, _allot
		ldr	TMP0, [SP], #4
		and	TOS, TMP0, TOS
		NEXT
	CODE_END

	VARIABLE	0, "BASE", _base, _and

	CODE_BEGIN	IMMEDIATE, "BEGIN", _begin, _base
		str	HERE, [CP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "BL", _bl, _begin
		str	TOS, [SP, #-4]!
		mov	TOS, #' '
		NEXT
	CODE_END

	CODE_BEGIN	0, "BYE", _bye, _bl
		b	gbye
	CODE_END

	CODE_BEGIN	0, "C!", _c_store, _bye
		ldmfd	SP!, {TMP0, TMP1}
		strb	TMP0, [TOS]
		mov	TOS, TMP1
		NEXT
	CODE_END

	CODE_BEGIN	0, "C,", _c_comma, _c_store
		strb	TOS, [HERE], #1
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "C@", _c_fetch, _c_comma
		ldrb	TOS, [TOS]
		NEXT
	CODE_END

	CODE_BEGIN	0, "CELL+", _cell_plus, _c_fetch
		add	TOS, TOS, #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "CELL-", _cell_minus, _cell_plus
		sub	TOS, TOS, #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "CELLS", _cells, _cell_minus
		mov	TOS, TOS, lsl #2
		NEXT
	CODE_END

	CODE_BEGIN	0, "CHAR+", _char_plus, _cells
		add	TOS, TOS, #1
		NEXT
	CODE_END

	CODE_BEGIN	0, "CHAR-", _char_minus, _char_plus
		sub	TOS, TOS, #1
		NEXT
	CODE_END

	ALIAS	0, "CHARS", _chars, _nop, _char_minus

	/* ANS says this is a STRING word, but I don't agree (see FIND) */
	CODE_BEGIN	0, "COMPARE", _compare, _chars
		ldmfd	SP!, {TMP0, TMP1, TMP2}
1:		teq	TOS, #0
		teqne	TMP1, #0
		beq	2f
		ldrb	TMP3, [TMP0], #1
		ldrb	TMP4, [TMP2], #1
		subs	TMP3, TMP4, TMP3
		bne	3f
		sub	TOS, TOS, #1
		sub	TMP1, TMP1, #1
		b	1b
2:		cmp	TMP1, TOS
3:		movhi	TMP3, #1
		movlo	TMP3, #-1
		moveq	TMP3, #0
		mov	TOS, TMP3
		NEXT
	CODE_END

	CODE_BEGIN	0, "COUNT", _count, _compare
		ldrb	TMP0, [TOS], #1
		str	TOS, [SP, #-4]!
		mov	TOS, TMP0
		NEXT
	CODE_END

	CODE_BEGIN	0, "DEPTH", _depth, _count
		ldr	TMP0, [UP, #_off_sp0]
		sub	TMP0, TMP0, SP
		str	TOS, [SP, #-4]!
		mov	TOS, TMP0, asr #2
		NEXT
	CODE_END

	/* Implementation specific - runtime for DO */
	CODE_BEGIN	COMPILE_ONLY, "(DO)", _do_do, _depth
		ldmfd	SP!, {TMP0, TMP1}
		rsb	TMP0, TMP0, #(1 << 31)
		add	TOS, TOS, TMP0
		stmfd	RP!, {TOS, TMP0}
		mov	TOS, TMP1
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "DO", _do, _do_do
		ldr	TMP0, =_do_do
		str	TMP0, [HERE], #4
		str	HERE, [CP, #-4]!	/* Start of loop */
		str	TOS, [SP, #-4]!
		mov	TOS, #0			/* LEAVE fixups */
		NEXT
	CODE_END

	CODE_BEGIN	0, "DROP", _drop, _do
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "DUP", _dup, _drop
		str	TOS, [SP, #-4]!
		NEXT
	CODE_END

	/* Implementation specific - runtime for ELSE */
	CODE_BEGIN	COMPILE_ONLY, "BRANCH", _branch, _dup
		ldr	TMP0, [IP, #4]
		sub	IP, TMP0, #4
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "ELSE", _else, _branch
		ldr	TMP0, =_branch
		str	TMP0, [HERE], #4
		swp	TMP0, HERE, [CP]
		add	HERE, HERE, #4
		str	HERE, [TMP0]
		NEXT
	CODE_END

	CODE_BEGIN	0, "ENVIRONMENT?", _environment_question, _else
			CSTRING	"/COUNTED-STRING"
			.align	2 ;	.long	255
			CSTRING	"/HOLD"
			.align	2 ;	.long	HOLD_SIZE
			CSTRING	"/PAD"
			.align	2 ;	.long	PAD_SIZE
			CSTRING	"ADDRESS-UNIT-BITS"
			.align	2 ;	.long	32
			CSTRING	"FLOORED"
			.align	2 ;	.long	FALSE
			CSTRING	"MAX-CHAR"
			.align	2 ;	.long	255
			CSTRING	"RETURN-STACK-CELLS"
			.align	2 ;	.long	(RP_SIZE / 4)
			CSTRING	"STACK-CELLS"
			.align	2 ;	.long	(SP_SIZE / 4)
			CSTRING	"MAX-N"
			.align	2 ;	.long	0x7fffffff
			CSTRING	"MAX-U"
			.align	2 ;	.long	0xffffffff

		#if defined(CORE_EXT)
			CSTRING	"CORE-EXT"
			.align	2 ; 	.long	1
		#endif
		#if defined(BLOCK)
			CSTRING	"BLOCK"
			.align	2 ; 	.long	1
		#endif
		#if defined(BLOCK_EXT)
			CSTRING	"BLOCK-EXT"
			.align	2 ; 	.long	1
		#endif
		#if defined(DOUBLE)
			CSTRING	"DOUBLE"
			.align	2 ; 	.long	1
		#endif
		#if defined(DOUBLE_EXT)
			CSTRING	"DOUBLE-EXT"
			.align	2 ; 	.long	1
		#endif
		#if defined(FACILITY)
			CSTRING	"FACILITY"
			.align	2 ; 	.long	1
		#endif
		#if defined(FACILITY_EXT)
			CSTRING	"FACILITY-EXT"
			.align	2 ; 	.long	1
		#endif
		#if defined(FILE)
			CSTRING	"FILE"
			.align	2 ; 	.long	1
		#endif
		#if defined(FILE_EXT)
			CSTRING	"FILE-EXT"
			.align	2 ; 	.long	1
		#endif
		#if defined(LOCAL)
			CSTRING	"LOCAL"
			.align	2 ; 	.long	1
		#endif
		#if defined(LOCAL_EXT)
			CSTRING	"LOCAL-EXT"
			.align	2 ; 	.long	1
		#endif
		#if defined(TOOLS)
			CSTRING	"TOOLS"
			.align	2 ; 	.long	1
		#endif
		#if defined(TOOLS_EXT)
			CSTRING	"TOOLS-EXT"
			.align	2 ; 	.long	1
		#endif
		#if defined(SEARCH)
			CSTRING	"SEARCH"
			.align	2 ; 	.long	1
		#endif
		#if defined(SEARCH_EXT)
			CSTRING	"SEARCH-EXT"
			.align	2 ; 	.long	1
		#endif
		#if defined(STRING)
			CSTRING	"STRING"
			.align	2 ; 	.long	1
		#endif
		#if defined(STRING_EXT)
			CSTRING	"STRING-EXT"
			.align	2 ; 	.long	1
		#endif
		.byte	0
	CODE_END

	CODE_BEGIN	0, "EXECUTE", _execute, _environment_question
		mov	TMP0, TOS
		ldr	TOS, [SP], #4
		ldr	pc, [TMP0]
	CODE_END

	CODE_BEGIN	0, "EXIT", _exit, _execute
		ldr	IP, [RP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "FILL", _fill, _exit
		ldmfd	SP!, {TMP0, TMP1, TMP2}

		orr	TOS, TOS, TOS, lsl #8
		orr	TOS, TOS, TOS, lsl #16

		teq	TMP0, #0
_fill1:		beq	8f
		tst	TMP1, #3
		beq	2f
		strb	TOS, [TMP1], #1
		subs	TMP0, TMP0, #1
		b	_fill1

2:		cmp	TMP0, #64
		blo	3f
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		subs	TMP0, TMP0, #64
		b	2b

3:		cmp	TMP0, #32
		blo	4f
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		subs	TMP0, TMP0, #32
		b	3b

4:		cmp	TMP0, #16
		blo	5f
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		subs	TMP0, TMP0, #16
		b	4b

5:		cmp	TMP0, #8
		blo	6f
		str	TOS, [TMP1], #4
		str	TOS, [TMP1], #4
		subs	TMP0, TMP0, #8
		b	5b

6:		cmp	TMP0, #4
		blo	7f
		str	TOS, [TMP1], #4
		subs	TMP0, TMP0, #4

7:		beq	8f
		strb	TOS, [TMP1], #4
		subs	TMP0, TMP0, #1
		b	7b

8:		mov	TOS, TMP2
		NEXT
	CODE_END

	CODE_BEGIN	0, "HERE", _here, _fill
		str	TOS, [SP, #-4]!
		mov	TOS, HERE
		NEXT
	CODE_END

	CODE_BEGIN	0, "I", _i, _here
		ldmfd	RP, {TMP0, TMP1}
		str	TOS, [SP, #-4]!
		sub	TOS, TMP0, TMP1
		NEXT
	CODE_END

	/* Implementation specific - runtime for IF */
	CODE_BEGIN	COMPILE_ONLY, "?BRANCH", _question_branch, _i
		teq	TOS, #FALSE
		ldr	TOS, [SP], #4
		ldr	TMP0, [IP, #4]!
		subeq	IP, TMP0, #4
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "IF", _if, _question_branch
		ldr	TMP0, =_question_branch
		str	TMP0, [HERE], #4
		str	HERE, [CP, #-4]!
		add	HERE, HERE, #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "INVERT", _invert, _if
		mvn	TOS, TOS
		NEXT
	CODE_END

	CODE_BEGIN	0, "J", _j, _invert
		add	TMP2, SP, #8
		ldmfd	TMP2, {TMP0, TMP1}
		str	TOS, [SP, #-4]!
		sub	TOS, TMP0, TMP1
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "LEAVE", _leave, _j
		ldr	TMP0, =_unloop
		str	TMP0, [HERE], #4
		ldr	TMP0, =_branch
		str	TMP0, [HERE], #4
		add	TOS, TOS, #1
		str	HERE, [SP, #-4]!
		add	HERE, HERE, #4
		NEXT
	CODE_END

	/* Implementation specific - runtime for LITERAL */
	CODE_BEGIN	COMPILE_ONLY, "LIT", _lit, _leave
		str	TOS, [SP, #-4]!
		ldr	TOS, [IP, #4]!
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "LITERAL", _literal, _lit
		ldr	TMP0, =_lit
		str	TMP0, [HERE], #4
		str	TOS, [HERE], #4
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	/* Implementation specific - runtime for LOOP */
	CODE_BEGIN	COMPILE_ONLY, "(LOOP)", _do_loop, _literal
		ldmfd	RP!, {TMP0, TMP1}
		adds	TMP0, TMP0, #1
		NEXT	vs, 8
		stmfd	RP!, {TMP0, TMP1}
		ldr	IP, [IP, #4]
		NEXT	al, 0
	CODE_END

	CODE_BEGIN	IMMEDIATE, "LOOP", _loop, _do_loop
		ldr	TMP0, =_do_loop
		str	TMP0, [HERE], #4
		ldr	TMP0, [CP], #4
		str	TMP0, [HERE], #4
1:		teq	TOS, #0
		ldreq	TOS, [SP], #4
		NEXT	eq
		ldr	TMP0, [SP], #4
		str	HERE, [TMP0]
		subs	TOS, TOS, #1
		b	1b
	CODE_END

	CODE_BEGIN	0, "LSHIFT", _lshift, _loop
		ldr	TMP0, [SP], #4
		mov	TOS, TMP0, lsl TOS
		NEXT
	CODE_END

	CODE_BEGIN	0, "M*", _m_star, _lshift
		ldr	TMP1, [SP], #4
		smull	TMP0, TOS, TMP1, TOS
		str	TMP0, [SP, #-4]!
		NEXT
	CODE_END

	/* ANS says this is a CORE-EXT word, but I disagree (see >NUMBER) */
	CODE_BEGIN	0, "M+", _m_plus, _m_star
		ldmfd	SP!, {TMP0, TMP1}
		adds	TMP1, TMP1, TOS
		adc	TOS, TMP0, #0
		str	TMP1, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "MAX", _max, _m_plus
		ldr	TMP0, [SP], #4
		cmp	TMP0, TOS
		movgt	TOS, TMP0
		NEXT
	CODE_END

	CODE_BEGIN	0, "MIN", _min, _max
		ldr	TMP0, [SP], #4
		cmp	TMP0, TOS
		movlt	TOS, TMP0
		NEXT
	CODE_END

	CODE_BEGIN	0, "MOVE", _move, _min
		ldm	SP!, {TMP0, TMP1}

		mov	TOS, TOS, lsl #2
		teq	TOS, #0
		beq	6f

		cmp	TMP0, TMP1
		blo	3f

		tst	TMP1, #3
		bne	2f
		tst	TMP0, #3
		bne	2f

1:		subs	TOS, TOS, #4
		ldr	TMP2, [TMP1, +TOS]
		str	TMP2, [TMP0, +TOS]
		bne	1b
		b	6f

2:		subs	TOS, TOS, #1
		ldrb	TMP2, [TMP1, +TOS]
		strb	TMP2, [TMP0, +TOS]
		bne	2b
		b	6f

3:		tst	TMP1, #3
		bne	5f
		tst	TMP0, #3
		bne	5f

4:		ldr	TMP2, [TMP1], #4
		str	TMP2, [TMP1], #4
		subs	TOS, TOS, #4
		bne	4b
		b	6f

5:		ldrb	TMP2, [TMP1], #1
		strb	TMP2, [TMP0], #1
		subs	TOS, TOS, #1
		bne	5b

6:		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "NEGATE", _negate, _move
		rsb	TOS, TOS, #0
		NEXT
	CODE_END

	CODE_BEGIN	0, "OR", _or, _negate
		ldr	TMP0, [SP], #4
		orr	TOS, TMP0, TOS
		NEXT
	CODE_END

	CODE_BEGIN	0, "OVER", _over, _or
		str	TOS, [SP, #-4]!
		ldr	TOS, [SP, #4]
		NEXT
	CODE_END

	CODE_BEGIN	0, "R>", _r_from, _over
		str	TOS, [SP, #-4]!
		ldr	TOS, [RP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "R@", _r_fetch, _r_from
		str	TOS, [SP, #-4]!
		ldr	TOS, [RP]
		NEXT
	CODE_END

	CODE_BEGIN	0, "RECURSE", _recurse, _r_fetch
		ldr	TMP0, =_var_last
		ldr	TMP0, [TMP0]
		ldrb	TMP1, [TMP0], #1
		bic	TMP0, TMP0, #(IMMEDIATE | COMPILE_ONLY)
		add	TMP0, TMP0, TMP1
		add	TMP0, TMP0, #11
		bic	TMP0, TMP0, #3
		mov	TMP1, HERE
1:		cmp	TMP0, TMP1
		ldrlo	TMP2, [TMP0], #4
		strlo	TMP2, [HERE], #4
		blo	1b
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "REPEAT", _repeat, _recurse
		ldmfd	CP!, {TMP0, TMP1}
		ldr	TMP2, =_branch
		str	TMP2, [HERE], #4
		str	TMP0, [HERE], #4
		str	HERE, [TMP1]
		NEXT
	CODE_END

	CODE_BEGIN	0, "ROT", _rot, _repeat
		swp	TOS, TOS, [SP]
		add	TMP0, SP, #4
		swp	TOS, TOS, [TMP0]
		NEXT
	CODE_END

	CODE_BEGIN	0, "RSHIFT", _rshift, _rot
		ldr	TMP0, [SP], #4
		mov	TOS, TMP0, lsr TOS
		NEXT
	CODE_END

	CODE_BEGIN	0, "S>D", _s_to_d, _rshift
		tst	TOS, #(1 << 31)
		str	TOS, [SP, #-4]!
		mov	TOS, #0
		movne	TOS, #-1
		NEXT
	CODE_END

	CODE_BEGIN	0, "SM/REM", _sm_slash_rem, _s_to_d
		/* Make divisor into a double */
		movs	TMP0, TOS
		movpl	TOS, #0
		movmi	TOS, #-1
		/* Load the dividend */
		ldmfd	SP!, {TMP1, TMP2}
		/* Divide */
		bl	divide64
		/* Keep quotient low and remainder low */
		mov	TOS, TMP2
		str	TMP4, [SP, #-4]!
		NEXT
	COLON_END

	VARIABLE	0, "STATE", _state, _sm_slash_rem

	CODE_BEGIN	0, "SWAP", _swap, _state
		swp	TOS, TOS, [SP]
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "THEN", _then, _swap
		ldr	TMP0, [CP], #4
		str	HERE, [TMP0]
		NEXT
	CODE_END

	CODE_BEGIN	0, "U<", _u_less, _then
		ldr	TMP0, [SP], #4
		cmp	TMP0, TOS
		mov	TOS, #FALSE
		movlo	TOS, #TRUE
		NEXT
	CODE_END

	CODE_BEGIN	0, "UM*", _um_star, _u_less
		ldr	TMP1, [SP], #4
		umull	TMP0, TOS, TMP1, TOS
		str	TMP0, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "UM/MOD", _um_slash_mod, _um_star
		/* Make divisor into a double */
		movs	TMP0, TOS
		mov	TOS, #0
		/* Load the dividend */
		ldmfd	SP!, {TMP1, TMP2}
		/* Divide */
		bl	udivide64
		/* Keep quotient low and remainder low */
		mov	TOS, TMP2
		str	TMP4, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "UNLOOP", _unloop, _um_slash_mod
		add	RP, RP, #8
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "UNTIL", _until, _unloop
		ldr	TMP0, =_question_branch
		str	TMP0, [HERE], #4
		ldr	TMP0, [CP], #4
		str	TMP0, [HERE], #4
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "WHILE", _while, _until
		ldr	TMP0, =_question_branch
		str	TMP0, [HERE], #4
		swp	TMP0, HERE, [CP]
		add	HERE, HERE, #4
		str	TMP0, [CP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "XOR", _xor, _while
		ldr	TMP0, [SP], #4
		eor	TOS, TMP0, TOS
		NEXT
	CODE_END

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Core words (colon definitions)
 */

	/* Implementation specific - double unsigned divide with remainder */
	CODE_BEGIN	0, "DU/MOD", _du_slash_mod, _xor
		/* Load divisor low and dividend */
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		/* Divide */
		bl	udivide64
		/* Keep quotient and remainder */
		mov	TOS, TMP1
		stmfd	SP!, {TMP2, TMP3, TMP4}
		NEXT
	CODE_END

	COLON_BEGIN	0, "#", _number, _du_slash_mod
		/* Make BASE a double cell */
		.long	_base, _fetch, _lit, 0
		/* Divide by BASE */
		.long	_du_slash_mod
		/* Throw away the high cell of the remainder */
		.long	_2swap, _drop
		/* Get offset for a numeric digit (0-9) */
		.long	_lit, 9, _over, _less
		/* Get offset for an alphabetic digit (A-Z) */
		.long	_lit, 7, _and, _plus
		/* Make digit an ASCII character */
		.long	_lit, '0', _plus
		/* Save character and exit */
		.long	_hold
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "#>", _end_picture, _number
		.long	_2drop
		.long	_holdp, _fetch
		.long	_hold0, _fetch
		.long	_over
		.long	_minus
		.long	_exit
	COLON_END

	/*
	 *	: #S
	 *		BEGIN
	 *			#
	 *			2DUP OR 0=
	 *		UNTIL
	 *	;
	 */
	COLON_BEGIN	0, "#S", _number_s, _end_picture
		/* Convert another digit */
1:		.long	_number
		/* Check if double cell equals zero */
		.long	_2dup, _or, _0_equal
		.long	_question_branch, 1b
			.long	_exit
	COLON_END

	COLON_BEGIN	0, "'", _tick, _number_s
		.long	_bl, _word, _dup, _c_fetch
		.long	_question_branch, 1f
			.long	_find
			.long	_question_branch, 1f
				.long	_exit
1:		.long	_drop
		.long	_lit, 0, _display_error
		/* Not reached */
	COLON_END

	COLON_BEGIN	0, "(", _paren, _tick
		.long	_lit, ')', _word
		.long	_drop
		.long	_exit
	COLON_END

	COLON_BEGIN	IMMEDIATE_ONLY, ".\"", _display_quote, _paren
		.long	_s_quote
		.long	_lit, _type, _comma
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "<#", _begin_picture, _display_quote
		.long	_here
		.long	_lit, HOLD_SIZE
		.long	_plus
		.long	_dup
		.long	_hold0
		.long	_store
		.long	_holdp
		.long	_store
		.long	_exit
	COLON_END

	/* Implementation specifc - unsigned to string */
	COLON_BEGIN	0, "U>$", _u_to_string, _begin_picture
		.long	_begin_picture
		.long	_lit, 0
		.long	_number_s
		.long	_end_picture
		.long	_exit
	COLON_END

	/* Implementation specifc - number to string */
	COLON_BEGIN	0, ">$", _to_string, _u_to_string
		.long	_base
		.long	_fetch
		.long	_lit, 10
		.long	_equal
		.long	_question_branch, 1f
			.long	_begin_picture
			.long	_dup
			.long	_abs
			.long	_s_to_d
			.long	_number_s
			.long	_rot
			.long	_sign
			.long	_end_picture
			.long	_exit
1:		.long	_u_to_string
		.long	_exit
	COLON_END

	COLON_BEGIN	0, ".", _display, _to_string
		.long	_to_string
		.long	_type
		.long	_exit
	COLON_END

	/* Implementation specific - convert a character to a number */
	COLON_BEGIN	0, "DIGIT?", _digit_question, _display
		.long	_lit, '0'
		.long	_minus
		.long	_lit, 9
		.long	_over
		.long	_less
		.long	_question_branch, 1f
			.long	_lit, 7
			.long	_minus
			.long	_dup
			.long	_lit, 10
			.long	_less
			.long	_or
1:		.long	_dup
		.long	_base
		.long	_fetch
		.long	_u_less
		.long	_exit
	COLON_END
		
	COLON_BEGIN	0, ">NUMBER", _to_number, _digit_question
		/* Check for 0 length string */
1:		.long	_dup
		.long	_lit, 0
		.long	_greater
		.long	_question_branch, 3f
			/* Validate the next digit */
			.long	_over
			.long	_c_fetch
			.long	_digit_question
			.long	_question_branch, 2f
				.long	_to_r
				.long	_2swap
				/* Multiply double cell by BASE */
				.long	_base
				.long	_fetch
				.long	_um_star
				.long	_drop
				.long	_swap
				.long	_base
				.long	_fetch
				.long	_um_star
				.long	_rot
				.long	_plus
				/* Add in the character */
				.long	_r_from
				.long	_m_plus
				/* Update the count and address */
				.long	_2swap
				.long	_1_minus
				.long	_swap
				.long	_1_plus
				.long	_swap
				.long	_branch, 1b
2:			.long	_drop
3:		.long	_exit
	COLON_END

	COLON_BEGIN	0, "ABORT", _abort, _to_number
		.long	_sp0, _fetch, _sp_store
		.long	_quit
		/* Not reached */
	COLON_END

	COLON_BEGIN	IMMEDIATE_ONLY, "ABORT\"", _abort_quote, _abort
		.long	_lit, _question_branch, _comma
		.long	_here, _lit, 0, _comma		/* Branch address */
		.long	_display_quote
		.long	_lit, _cr, _comma
		.long	_lit, _abort, _comma
		.long	_here, _swap, _store		/* Fixup branch */
		.long	_exit
	COLON_END

	/* Implementation specific - input a line */
	VARIABLE	0, "`ACCEPT", _tick_accept, _abort_quote

	COLON_BEGIN	0, "ACCEPT", _accept, _tick_accept
		.long	_tick_accept, _fetch, _execute
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "CHAR", _char, _accept
		.long	_bl, _word, _count
		.long	_question_branch, 1f
			.long	_c_fetch
			.long	_exit
1:		.long	_drop
		.long	_lit, 0, _display_error
		/* Not reached */
	COLON_END

	/* Implementation specific - runtime for CONSTANT */
	CODE_BEGIN	COMPILE_ONLY, "(CONSTANT)", _do_constant, _char
		str	TOS, [SP, #-4]!
		ldr	TOS, [TMP0, #4]
		NEXT
	CODE_END

	COLON_BEGIN	IMMEDIATE, "CONSTANT", _constant, _do_constant
		.long	_bl, _word, _count, _dup
		.long	_question_branch, 1f
			.long	_align, _here
			.long	_last
			.long	_dup, _fetch, _comma
			.long	_lit, 0, _c_comma
			.long	_rot, _char_plus, _allot
			.long	_align
			.long	_lit, _code_do_constant, _comma
			.long	_store
			.long	_drop
			.long	_comma
			.long	_exit
1:		.long	_2drop
		.long	_lit, 0, _display_error
		/* Not reached */
	COLON_END

	/* Implementation specific - output new line */
	VARIABLE	0, "`CR", _tick_cr, _constant

	COLON_BEGIN	0, "CR", _cr, _tick_cr
		.long	_tick_cr, _fetch, _execute
		.long	_exit
	COLON_END

	/* Implementation specific - runtime for words created with CREATE */
	CODE_BEGIN	COMPILE_ONLY, "(CREATE)", _do_create, _cr
		str	IP, [RP, #-4]!		/* Enter new thread */
		ldr	IP, [TMP0, #4]
		str	TOS, [SP, #-4]!		/* Push parameter field addr */
		add	TOS, TMP0, #8
		NEXT	al, 0
	CODE_END

	COLON_BEGIN	0, "CREATE", _create, _do_create
		.long	_bl, _word, _count, _dup
		.long	_question_branch, 2f
			.long	_align
			.long	_here, _last
			.long	_dup, _fetch, _comma	/* Link */
			.long	_store			/* Add to dictionary */
			.long	_lit, 0, _c_comma	/* Flags */
			.long	_char_plus, _allot	/* CSTRING */
			.long	_here, _lit, 3, _lit, 0, _fill
			.long	_align
			.long	_lit, _code_do_create, _comma
			.long	_lit, 1f, _comma	/* Thread address */
			.long	_drop
1:			.long	_exit
2:		.long	_2drop
		.long	_lit, 0, _display_error
		/* Not reached */
	COLON_END

	COLON_BEGIN	0, "DECIMAL", _decimal, _create
		.long	_lit, 10, _base, _store
		.long	_exit
	COLON_END

	COLON_BEGIN	COMPILE_ONLY, "(DOES>)", _do_does, _decimal
		.long	_last, _fetch
		.long	_cell_plus, _char_plus
		.long	_dup, _c_fetch, _plus, _char_plus, _aligned
		.long	_cell_plus
		.long	_store
		.long	_exit
	COLON_END

	COLON_BEGIN	IMMEDIATE_ONLY, "DOES>", _does, _do_does
		.long	_here, _lit, 4, _cells, _plus, _literal
		.long	_lit, _do_does, _comma
		.long	_lit, _exit, _comma
		.long	_exit
	COLON_END

	/* Implementation specific - method to change output device */
	VARIABLE	0, "`EMIT", _tick_emit, _does

	COLON_BEGIN	0, "EMIT", _emit, _tick_emit
		.long	_tick_emit, _fetch, _execute
		.long	_exit
	COLON_END

	/* Implementation specific - display an error message */
	COLON_BEGIN	0, ".ERROR", _display_error, _emit
		.long	_lit, _cstr_err_error, _count
		.long	_type
		.long	_cells
		.long	_lit, _err_tbl
		.long	_fetch
		.long	_count
		.long	_type
		.long	_cr
		.long	_quit
		/* Not reached */
		CSTRING	" ERROR: ", _err_error
		CSTRING "zero length string", _err_0
		/* Error table */
		.align 2
_err_tbl:	.long	_cstr_err_0
	COLON_END

	/* Implementation specific - start and end of stack used by HOLD */
	VARIABLE	0, "HOLD0", _hold0, _display_error
	VARIABLE	0, "HOLDP", _holdp, _hold0

	COLON_BEGIN	0, "HOLD", _hold, _holdp
		.long	_lit, -1, _holdp, _plus_store
		.long	_holdp
		.long	_fetch, _c_store
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "IMMEDIATE", _immediate, _hold
		.long	_last, _fetch
		.long	_cell_plus, _dup, _c_fetch
		.long	_lit, IMMEDIATE, _or
		.long	_swap, _c_store
		.long	_exit
	COLON_END

	/* Implementation specific - method to change input device */
	VARIABLE	0, "`KEY", _tick_key, _immediate

	COLON_BEGIN	0, "KEY", _key, _tick_key
		.long	_tick_key, _fetch, _execute
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "MOD", _mod, _key
		.long	_slash_mod, _drop
		.long	_exit
	COLON_END
	
	COLON_BEGIN	IMMEDIATE, "POSTPONE", _postpone, _mod
		.long	_bl, _word, _dup, _c_fetch
		.long	_question_branch, 1f
			.long	_tick, _comma
			.long	_exit
1:		.long	_drop
		.long	_lit, 0, _display_error
	COLON_END

	/* Implementation specific */
	VARIABLE	0, "`PROMPT", _tick_prompt, _postpone

	/* Implementation specific */
	COLON_BEGIN	0, "doPROMPT", _doPROMPT, _tick_prompt
		.long	_lit, 1f, _count, _type, _cr
		.long	_exit
		/* Not reached */
1:		CSTRING	"  OK"
	COLON_END

	COLON_BEGIN	0, "PROMPT", _prompt, _doPROMPT
		.long	_state, _fetch
		.long	_question_branch, 1f
			.long	_exit
1:		.long	_tick_prompt, _fetch, _execute
		.long	_exit
	COLON_END

	/* Implementation specific - string constant */
	CODE_BEGIN	COMPILE_ONLY, "DOSTR", _dostr, _prompt
		str	TOS, [SP, #-4]!
		add	TMP0, IP, #5		/* Address */
		ldrb	TOS, [TMP0], #1		/* COUNT */
		str	TMP0, [SP, #-4]!
		add	IP, TMP0, TOS
		add	IP, IP, #3
		bic	IP, IP, #3
		NEXT	al, 0
	CODE_END

	COLON_BEGIN	IMMEDIATE_ONLY, "S\"", _s_quote, _dostr
		.long	_lit, '"', _word, _count
		.long	_dup
		.long	_question_branch, 1f
			.long	_lit, _dostr, _comma
			.long	_lit, 0, _c_comma	/* Unused byte */
			.long	_char_plus, _allot
			.long	_align
			.long	_drop
			.long	_exit
1:		.long	_2drop
		.long	_lit, 0
		.long	_display_error
		/* Not reached */
	COLON_END

	COLON_BEGIN	0, "SIGN", _sign, _s_quote
		.long	_0_less
		.long	_question_branch, 1f
			.long	_lit, '-', _hold
1:		.long	_exit
	COLON_END

	COLON_BEGIN	0, "SOURCE", _source, _sign
		.long	_tib, _fetch
		.long	_number_tib, _fetch
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "SPACE", _space, _source
		.long	_bl, _emit
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "SPACES", _spaces, _space
1:		.long	_dup
		.long	_lit, 0
		.long	_greater
		.long	_question_branch, 2f
			.long	_space
			.long	_1_minus
			.long	_branch, 1b
2:		.long	_drop
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "TYPE", _type, _spaces
		.long	_lit, 0
		.long	_do_question_do, 2f
1:			.long	_dup
			.long	_i
			.long	_plus
			.long	_c_fetch
			.long	_emit
		.long	_do_loop, 1b
2:		.long	_drop
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "U.", _u_display, _type
		.long	_u_to_string
		.long	_type
		.long	_exit
	COLON_END

	/* Implementation specific - runtime for VARIABLE */
	CODE_BEGIN	COMPILE_ONLY, "DOVAR", _dovar, _u_display
		str	TOS, [SP, #-4]!
		ldr	TOS, [TMP0, #4]
		add	TOS, TOS, UP
		NEXT
	CODE_END

	COLON_BEGIN	IMMEDIATE, "VARIABLE", _variable, _dovar
		.long	_bl, _word, _count, _dup
		.long	_question_branch, 1f
			.long	_align, _here
			.long	_last
			.long	_dup, _fetch, _comma
			.long	_lit, 0, _c_comma
			.long	_rot, _char_plus, _allot
			.long	_align
			.long	_lit, _code_dovar, _comma
			.long	_here, _up0, _fetch, _minus, _cell_plus, _comma
			.long	_lit, 0, _comma
			.long	_store
			.long	_drop
			.long	_exit
1:		.long	_2drop
		.long	_lit, 0, _display_error
		/* Not reached */
	COLON_END

/*
 *	: WORD
 *		>R
 *		HERE DUP
 *
 *		( Skip multiple delimiters )
 *		TIB @ DUP #TIB @ + SWAP >IN @ +
 *		BEGIN
 *			2DUP U< INVERT IF
 *				DUP C@ R@ =
 *			THEN
 *		WHILE
 *			CHAR+
 *		REPEAT
 *
 *		( Parse a word )
 *		BEGIN
 *			2DUP U< INVERT IF
 *				DUP C@ R@ = INVERT
 *			THEN
 *		WHILE
 *			DUP C@ >R
 *			CHAR+
 *			2SWAP CHAR+
 *			R> OVER C!
 *			2SWAP
 *		REPEAT
 *		R> DROP
 *		CHAR+ TIB @ - >IN !
 *		DROP
 *		OVER - OVER C!
 *	;
 */
	COLON_BEGIN	0, "WORD", _word, _variable
		.long	_to_r
		.long	_here, _aligned, _cell_plus, _char_plus, _dup
		.long	_tib, _fetch, _dup, _number_tib, _fetch, _plus
		.long	_swap, _from_in, _fetch, _plus
		/* Ignore leading delimiters */
1:		.long	_2dup, _u_less, _invert
		.long	_question_branch, 2f
			.long	_dup, _c_fetch
			.long	_r_fetch, _equal
		.long	_question_branch, 2f
			.long	_char_plus
			.long	_branch, 1b
		/* Parse a word */
2:		.long	_2dup, _swap, _u_less
		.long	_question_branch, 3f
			.long	_dup, _c_fetch
			.long	_r_fetch, _equal, _invert
		.long	_question_branch, 3f
			.long	_dup, _c_fetch, _to_r
			.long	_char_plus
			.long	_2swap, _char_plus
			.long	_r_from, _over, _c_store
			.long	_2swap
			.long	_branch, 2b
3:		.long	_r_from, _drop
		.long	_char_plus, _tib, _fetch, _minus, _from_in, _store
		.long	_drop
		.long	_over, _minus, _over, _c_store
		.long	_exit
	COLON_END

	COLON_BEGIN	IMMEDIATE_ONLY, "[", _left_bracket, _word
		.long	_lit, 0, _state, _store
		.long	_exit
	COLON_END

	COLON_BEGIN	IMMEDIATE_ONLY, "[']", _bracket_tick, _left_bracket
		.long	_lit, _lit, _comma
		.long	_tick, _comma
		.long	_exit
	COLON_END

	COLON_BEGIN	IMMEDIATE_ONLY, "[CHAR]", _bracket_char, _bracket_tick
		.long	_bl, _word, _count
		.long	_question_branch, 1f
			.long	_lit, _do_constant, _comma
			.long	_c_fetch, _comma
			.long	_exit
1:		.long	_lit, 0, _display_error
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "]", _right_bracket, _bracket_char
		.long	_lit, 1, _state, _store
		.long	_exit
	COLON_END

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Implementation specific
 */

	CODE_BEGIN	0, "U/MOD", _u_slash_mod, _right_bracket
		ldr	TMP0, [SP]
		mov	TMP1, #0
		mov	TMP2, #32
1:		movs	TMP0, TMP0, lsl #1
		adc	TMP1, TMP1, TMP1
		cmp	TMP1, TOS
		subhs	TMP1, TMP1, TOS
		orrhs	TMP0, TMP0, #1
		subs	TMP2, TMP2, #1
		bne	1b
		str	TMP1, [SP]
		mov	TOS, TMP0
		NEXT
	CODE_END

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Core word extension (optional)
 */

#if defined(CORE_EXT)

	COLON_BEGIN	0, ".R", _display_r, _u_slash_mod
		.long	_to_r
		.long	_to_string
		.long	_r_from
		.long	_over
		.long	_minus
		.long	_spaces
		.long	_type
		.long	_exit
	COLON_END

	CODE_BEGIN	0, "0<>", _0_not, _display_r
		teq	TOS, #0
		mov	TOS, #FALSE
		movne	TOS, #TRUE
		NEXT
	CODE_END

	CODE_BEGIN	0, "0>", _0_greater, _0_not
		cmp	TOS, #0
		mov	TOS, #FALSE
		movgt	TOS, #TRUE
		NEXT
	CODE_END

	CODE_BEGIN	0, "2>R", _2_to_r, _0_greater
		ldmfd	SP!, {TMP0, TMP1}
		stmfd	RP!, {TOS, TMP0}
		mov	TOS, TMP1
		NEXT
	CODE_END

	CODE_BEGIN	0, "2R>", _2_r_from, _2_to_r
		mov	TMP1, TOS
		ldmfd	RP!, {TOS, TMP0}
		stmfd	SP!, {TMP0, TMP1}
		NEXT
	CODE_END

	CODE_BEGIN	0, "2R@", _2_r_fetch, _2_r_from
		mov	TMP1, TOS
		ldmfd	RP, {TOS, TMP0}
		stmfd	SP!, {TMP0, TMP1}
		NEXT
	CODE_END

	CODE_BEGIN	0, "<>", _not_equal, _2_r_fetch
		ldr	TMP0, [SP], #4
		teq	TOS, TMP0
		mov	TOS, #FALSE
		movne	TOS, #TRUE
	CODE_END

	/* Implementation specific - runtime for DO */
	CODE_BEGIN	COMPILE_ONLY, "(?DO)", _do_question_do, _not_equal
		ldmfd	SP!, {TMP0, TMP1}
		teq	TOS, TMP0
		bne	1f
		mov	TOS, TMP1
		ldr	IP, [IP, #4]
		NEXT	al, 0
1:		rsb	TMP0, TMP0, #(1 << 31)
		add	TOS, TOS, TMP0
		stmfd	RP!, {TOS, TMP0}
		mov	TOS, TMP1
		NEXT	al, 8
	CODE_END

	CODE_BEGIN	IMMEDIATE, "?DO", _question_do, _do_question_do
		ldr	TMP0, =_do_question_do
		str	TMP0, [HERE], #4
		str	TOS, [SP, #-4]!
		str	HERE, [SP, #-4]!	/* ?DO Fixup */
		mov	TOS, #1			/* LEAVE fixups */
		add	HERE, HERE, #4
		str	HERE, [CP, #-4]!	/* Start of loop */
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "AGAIN", _AGAIN, _question_do
		ldr	TMP0, =_branch
		str	TMP0, [HERE], #4
		ldr	TMP0, [CP], #4
		str	TMP0, [HERE], #4
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "CASE", _case, _AGAIN
		ldr	TMP0, =_skip
		str	TMP0, [HERE], #4
		str	HERE, [CP, #-4]!
		add	HERE, HERE, #4
		NEXT
	CODE_END

	/* Implementation specific - runtime for OF */
	CODE_BEGIN	COMPILE_ONLY, "(OF)", _do_of, _case
		ldr	TMP0, [SP], #4
		teq	TOS, TMP0
		ldreq	TOS, [SP], #4
		NEXT	eq, 8
		mov	TOS, TMP0
		ldr	IP, [IP, #4]
		NEXT	al, 0
	CODE_END

	CODE_BEGIN	IMMEDIATE, "OF", _of, _do_of
		ldr	TMP0, =_do_of
		str	TMP0, [HERE], #4
		str	HERE, [CP, #-4]!
		add	HERE, HERE, #4
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "ENDOF", _endof, _of
		ldmfd	CP!, {TMP0, TMP1}
		str	HERE, [TMP1]		/* Resolve old case-sys */
		ldr	TMP2, =_branch
		str	TMP2, [HERE], #4
		str	HERE, [CP, #-4]!	/* New case-sys */
		add	HERE, HERE, #4
		str	HERE, [TMP0]		/* Resolve of-sys */
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "ENDCASE", _endcase, _endof
		ldr	TMP0, =_drop
		str	TMP0, [HERE], #4
		ldr	TMP0, [CP], #4
		str	HERE, [TMP0]
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "COMPILE,", _compile_comma, _endcase
		str	TOS, [HERE], #4
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "ERASE", _erase, _compile_comma
		ldmfd	SP!, {TMP0, TMP1}
		teq	TOS, #0
		beq	2f
		mov	TMP2, #0
1:		subs	TOS, TOS, #1
		strb	TMP2, [TMP0, TOS]
		bne	1b
2:		mov	TOS, TMP1
		NEXT
	CODE_END

	CODE_BEGIN	0, "FALSE", _false, _erase
		str	TOS, [SP, #-4]!
		mov	TOS, #FALSE
		NEXT
	CODE_END

	COLON_BEGIN	0, "HEX", _hex, _false
		.long	_lit, 16
		.long	_base, _store
		.long	_exit
	COLON_END

	CODE_BEGIN	0, "NIP", _nip, _hex
		add	SP, SP, #4
		NEXT
	CODE_END

	COLON_BEGIN	0, "PAD", _pad, _nip
		.long	_here
		.long	_lit, HOLD_SIZE
		.long	_plus
		.long	_exit
	COLON_END

	CODE_BEGIN	0, "PICK", _pick, _pad
		ldr	TOS, [SP, TOS, lsl #2]
		NEXT
	CODE_END

	CODE_BEGIN	0, "ROLL", _roll, _pick
		ldr	TMP0, [SP, TOS, lsl #2]
		sub	TMP1, TOS, #1
		teq	TOS, #0
1:		beq	2f
		ldr	TMP2, [SP, TMP1, lsl #2]
		str	TMP2, [SP, TOS, lsl #2]
		sub	TMP1, TMP1, #1
		subs	TOS, TOS, #1
		b	1b
2:		mov	TOS, TMP0
		add	SP, SP, #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "TRUE", _true, _roll
		str	TOS, [SP, #-4]!
		mov	TOS, #TRUE
		NEXT
	CODE_END

	CODE_BEGIN	0, "TUCK", _tuck, _true
		ldr	TMP0, [SP], #4
		mov	TMP1, TOS
		stmfd	SP!, {TMP0, TMP1}
		NEXT
	CODE_END

	CODE_BEGIN	0, "U>", _u_greater, _tuck
		ldr	TMP0, [SP], #4
		cmp	TMP0, TOS
		mov	TOS, #FALSE
		movhi	TOS, #TRUE
		NEXT
	CODE_END

	COLON_BEGIN	0, "UNUSED", _unused, _u_greater
		.long	_uptop, _fetch
		.long	_here
		.long	_minus
		.long	_exit
	COLON_END

	CODE_BEGIN	0, "WITHIN", _within, _unused
		ldmfd	SP!, {TMP0, TMP1}
		sub	TOS, TOS, TMP0
		sub	TMP1, TMP1, TMP0
		cmp	TMP1, TOS
		mov	TOS, #FALSE
		movlo	TOS, #TRUE
		NEXT
	CODE_END

#endif /* defined(CORE_EXT) */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#if defined(DOUBLE)
/*
 * Double cell (optional)
 */

	CODE_BEGIN	0, "D+", _d_plus, _within
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		adds	TMP0, TMP2, TMP0
		adc	TOS, TMP1, TOS
		str	TMP0, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "D-", _d_minus, _d_plus
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		subs	TMP0, TMP2, TMP0
		sbc	TOS, TMP1, TOS
		str	TMP0, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "D0<", _d0_less, _d_minus
		ldr	TMP0, [SP], #4
		cmp	TOS, #0
		cmpeq	TMP0, #0
		mov	TOS, #FALSE
		movlt	TOS, #TRUE
		NEXT	
	CODE_END

	CODE_BEGIN	0, "D0=", _d0_equal, _d0_less
		ldr	TMP0, [SP], #4
		teq	TOS, #0
		teqeq	TMP0, #0
		mov	TOS, #FALSE
		moveq	TOS, #TRUE
		NEXT	
	CODE_END

	CODE_BEGIN	0, "D2*", _d2_star, _d0_equal
		ldr	TMP0, [SP]
		movs	TMP0, TMP0, lsl #1
		adc	TOS, TOS, TOS
		str	TMP0, [SP]
		NEXT	
	CODE_END

	CODE_BEGIN	0, "D2/", _d2_slash, _d2_star
		ldr	TMP0, [SP]
		movs	TOS, TOS, asr #1
		mov	TMP0, TMP0, rrx
		str	TMP0, [SP]
		NEXT	
	CODE_END

	CODE_BEGIN	0, "D<", _d_less, _d2_slash
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		cmp	TMP1, TOS
		cmplt	TMP2, TMP0
		mov	TOS, #FALSE
		movlt	TOS, #TRUE
		NEXT	
	CODE_END

	CODE_BEGIN	0, "D=", _d_equal, _d_less
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		teq	TMP1, TOS
		teqeq	TMP2, TMP0
		mov	TOS, #FALSE
		moveq	TOS, #TRUE
		NEXT	
	CODE_END

	CODE_BEGIN	0, "D>S", _d_to_s, _d_equal
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	CODE_BEGIN	0, "DABS", _dabs, _d_to_s
		tst	TOS, #(1 << 31)
		beq	1f
		ldr	TMP0, [SP]
		rsbs	TMP0, TMP0, #0
		rsc	TOS, TOS, #0
		str	TMP0, [SP]
1:		NEXT
	CODE_END

	CODE_BEGIN	0, "DMAX", _dmax, _dabs
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		cmp	TMP1, TOS
		cmpeq	TMP2, TMP0
		movgt	TOS, TMP1
		movgt	TMP0, TMP2
		str	TMP0, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "DMIN", _dmin, _dmax
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		cmp	TMP1, TOS
		cmpeq	TMP2, TMP0
		movlt	TOS, TMP1
		movlt	TMP0, TMP2
		str	TMP0, [SP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	0, "DNEGATE", _dnegate, _dmin
		ldr	TMP0, [SP]
		rsbs	TMP0, TMP0, #0
		rsc	TOS, TOS, #0
		str	TMP0, [SP]
	CODE_END

/*
 * Colon Definitions
 */

	/* Implementation specific - unsigned double to string */
	COLON_BEGIN	0, "DU>$", _du_to_string, _dnegate
		.long	_begin_picture
		.long	_number_s
		.long	_end_picture
		.long	_exit
	COLON_END

	/* Implementation specific - double to string */
	COLON_BEGIN	0, "D>$", _d_to_string, _du_to_string
		.long	_base
		.long	_fetch
		.long	_lit, 10
		.long	_equal
		.long	_question_branch, 1f
			.long	_dup
			.long	_to_r
			.long	_dabs
			.long	_begin_picture
			.long	_number_s
			.long	_r_from
			.long	_sign
			.long	_end_picture
			.long	_exit
1:		.long	_du_to_string
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "D.", _d_display, _d_to_string
		.long	_d_to_string
		.long	_type
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "D.R", _d_display_r, _d_display
		.long	_to_r
		.long	_d_to_string
		.long	_r_from
		.long	_over
		.long	_minus
		.long	_spaces
		.long	_type
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "DU.", _du_display, _d_display_r
		.long	_du_to_string
		.long	_type
		.long	_exit
	COLON_END

#endif /* defined(DOUBLE) */

#if defined(DOUBLE_EXT)
/*
 * Double cell extension (optional)
 */

	CODE_BEGIN	0, "2ROT", _2rot, _du_display
		ldmfd	SP!, {TMP2, TMP3, TMP4}
		mov	TMP1, TOS
		ldmfd	SP!, {TOS, TMP0}
		stmfd	SP!, {TOS, TMP0, TMP1, TMP2, TMP3, TMP4}
	CODE_END

	CODE_BEGIN	0, "DU<", _du_less, _2rot
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		cmp	TMP1, TOS
		cmpeq	TMP2, TMP0
		mov	TOS, #FALSE
		movlo	TOS, #TRUE
		NEXT
	CODE_END

#endif /* defined(DOUBLE_EXT)


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#if defined(FIXED) && defined(DOUBLE) && defined(DOUBLE_EXT)

/*
 * Fixed point math (optional)
 */
	CODE_BEGIN	0, "D>F", _d_to_f, _du_less
		mov	TMP0, #0
		swp	TOS, TMP0, [SP]
		NEXT
	CODE_END

	ALIAS	0, "F!", _f_store, _2_store, _d_to_f

/*
 *	Given the 64 bit numbers AB and CD, where ABCD are 32 bits each:
 *
 *	AB * CD = (A*(2^32) + B*(2^0)) * (C*(2^32) + D*(2^0))
 *		= A*(2^32) * C*(2^32) + A*(2^32) * D*(2^0)
 *		+ B*(2^0) * C*(2^32) + B*(2^0) * D*(2^0)
 *		= A*C*(2^64) + A*D*(2^32) + B*C*(2^32) + B*D*(2^0)
 *		= A*C*(2^64) + (A*D + B*C)*(2^32) + B*D*(2^0)
 *
 *	Only need the middle 64 bits of the 128 bit product.
 */
	CODE_BEGIN	0, "F*", _du_star, _f_store
		ldmfd	SP!, {TMP0, TMP1, TMP2}
		/* Assume positive result */
		mov	TMP4, #0
		/* Make multiplier positive */
		tst	TOS, #(1 << 31)
		beq	1f
		rsbs	TMP0, TMP0, #0
		rsc	TOS, TOS, #0
		eor	TMP4, TMP4, #1
		/* Make multiplicand positive */
1:		tst	TMP1, #(1 << 31)
		beq	2f
		rsbs	TMP2, TMP2, #0
		rsc	TMP1, TMP1, #0
		eor	TMP4, TMP4, #1
		/* A*D (discard msw and free register with A) */
2:		umull	TOS, TMP3, TMP2, TOS
		/* B*C (discard msw and free register with C) */
		umull	TMP1, TMP3, TMP0, TMP1
		/* (A*D + B*C)*(2^32) */
		add	TOS, TOS, TMP1
		/* B*D (free registers with B and D) */
		umull	TMP0, TMP1, TMP2, TMP0
		/* (A*D + B*C)*(2^32) + B*D*(2^0) */
		add	TOS, TOS, TMP1
		/* Fix the sign of the product */
		teq	TMP4, #0
		beq	3f
		rsbs	TMP0, TMP0, #0
		rsc	TOS, TOS, #0
		/* Save the result */
3:		str	TMP0, [SP, #-4]!
		/* Done */
		NEXT
	CODE_END

	ALIAS	0, "F+", _f_plus, _d_plus, _du_star
	ALIAS	0, "F-", _f_minus, _d_minus, _f_plus
	ALIAS	0, "F0<", _f0_less, _d0_less, _f_minus
	ALIAS	0, "F0=", _f0_equal, _d0_equal, _f0_less
	ALIAS	0, "F<", _f_less, _d_less, _f0_equal

	CODE_BEGIN	0, "F>D", _f_to_d, _f_less
		str	TOS, [SP]
		mov	TOS, #0
		NEXT
	CODE_END

	ALIAS	0, "F@", _f_fetch, _2_fetch, _f_to_d
	ALIAS	0, "FALIGN", _falign, _align, _f_fetch
	ALIAS	0, "FALIGNED", _faligned, _aligned, _falign
	ALIAS	0, "FDROP", _fdrop, _2drop, _faligned
	ALIAS	0, "FDUP", _fdup, _2dup, _fdrop
	ALIAS	0, "FMAX", _fmax, _dmax, _fdup
	ALIAS	0, "FMIN", _fmin, _dmin, _fmax
	ALIAS	0, "FNEGATE", _fnegate, _dnegate, _fmin
	ALIAS	0, "FOVER", _fover, _2over, _fnegate
	ALIAS	0, "FROT", _frot, _2rot, _fover
	ALIAS	0, "FSWAP", _fswap, _2swap, _frot

#endif /* defined(FIXED) && defined(DOUBLE) && defined(DOUBLE_EXT) */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#if defined (TOOLS)
/*
 * Programming Tools (optional)
 */

	/* Implementation specific */
	COLON_BEGIN	0, ">HEXDIGIT", _to_hexdigit, _fswap
		.long	_lit, 0xf
		.long	_and
		.long	_lit, 1f
		.long	_plus
		.long	_c_fetch
		.long	_exit
		CSTRING	"0123456789abcdef"
	COLON_END

	/* Implementation specific */
	COLON_BEGIN	0, ">NYBBLES", _to_nybbles, _to_hexdigit
		.long	_lit, 8
		.long	_lit, 1
		.long	_do_do
1:			.long	_dup
			.long	_lit, 4
			.long	_rshift
		.long	_do_loop, 1b
		.long	_exit
	COLON_END

	/* Implementation specific */
	COLON_BEGIN	0, ".HEX", _display_hex, _to_nybbles
		.long	_to_nybbles
		.long	_lit, 8
		.long	_lit, 0
		.long	_do_do
1:			.long	_to_hexdigit
			.long	_emit
		.long	_do_loop, 1b
		.long	_exit
	COLON_END

	COLON_BEGIN	0, ".S", _display_s, _display_hex
		.long	_sp_fetch
		.long	_sp0
		.long	_fetch
		.long	_cell_minus
		.long	_swap
		.long	_lit, '{', _emit
		.long	_base, _fetch
		.long	_depth, _lit, 3, _minus
		.long	_decimal, _display, _base, _store
		.long	_lit, '}', _emit
		.long	_do_question_do, 4f
1:			.long	_i, _fetch
			.long	_cr
			.long	_dup, _display
			.long	_dup
			.long	_bl, _lit, ('~' + 1), _within
			.long	_question_branch, 2f
				.long	_space
				.long	_emit
			.long	_branch, 3f
2:			.long	_drop
3:			.long	_lit, 4, _do_plus_loop, 1b
4:		.long	_exit
	COLON_END

	COLON_BEGIN	0, "?", _question, _display_s
		.long	_fetch
		.long	_display_hex
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "DUMP", _dump, _question
		.long	_over
		.long	_plus
		.long	_swap
		.long	_do_do
1:			.long	_i
			.long	_display_hex
			.long	_lit, ':'
			.long	_emit
			.long	_space
			.long	_i
			.long	_lit, 16
			.long	_plus
			.long	_i
			.long	_do_do
2:				.long	_i
				.long	_c_fetch
				.long	_dup
				.long	_to_hexdigit
				.long	_swap
				.long	_lit, 4
				.long	_rshift
				.long	_to_hexdigit
				.long	_emit
				.long	_emit
				.long	_space
			.long	_do_loop, 2b
			.long	_i
			.long	_lit, 16
			.long	_plus
			.long	_i
			.long	_do_do
4:				.long	_i
				.long	_c_fetch
				.long	_dup
				.long	_lit, ' '
				.long	_lit, ('~' + 1)
				.long	_within
				.long	_question_branch, 5f
					.long	_emit
					.long	_branch, 6f
5:				.long	_drop
				.long	_lit, '.'
				.long	_emit
6:			.long	_do_loop, 4b
			.long	_cr
			.long	_lit, 16
		.long	_do_plus_loop, 1b
8:		.long	_exit
	COLON_END

/*
 * : WORDS
 *	LAST
 *	BEGIN
 *		@ DUP
 *	WHILE
 *		DUP CELL+ CHAR+
 *		COUNT TYPE SPACE
 *	REPEAT
 *	CR CR
 *	DROP
 * ;
 */
	COLON_BEGIN	IMMEDIATE, "WORDS", _words, _dump
		.long	_last
1:		.long	_fetch, _dup
		.long	_question_branch, 2f
			.long	_dup, _cell_plus, _char_plus
			.long	_count, _type, _space
			.long	_branch, 1b
2:			.long	_cr, _cr 
			.long	_drop
			.long	_exit
	COLON_END

#endif /* defined(TOOLS) */
 
#if defined(TOOLS_EXT)
 
	CODE_BEGIN	IMMEDIATE, "AHEAD", _ahead, _words
		ldr	TMP0, =_skip
		str	TMP0, [HERE], #8
		str	HERE, [CP, #-4]!
		NEXT
	CODE_END

	CODE_BEGIN	IMMEDIATE, "CS-PICK", _cs_pick, _ahead
		ldr	TOS, [CP, TOS, lsl #2]
		str	TOS, [CP, #-4]!
		ldr	TOS, [SP], #4
		NEXT
	CODE_END

	ALIAS	IMMEDIATE, "[THEN]", _b_then, _nop, _cs_pick

#endif /* defined(TOOLS_EXT) */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#if defined(STRING)
/*
 * String words (optional)
 */

	CODE_BEGIN	0, "-TRAILING", _minus_trailing, _b_then
		ldr	TMP0, [SP]
		mov	TMP1, TOS
		teq	TMP1, #0
		beq	2f
1:		subs	TMP1, TMP1, #1
		blo	2f
		ldr	TMP2, [TMP0, TMP1]
		teq	TMP2, #32
		subeq	TOS, TOS, #1
		beq	1b
2:		NEXT
	CODE_END

	CODE_BEGIN	0, "/STRING", _slash_string, _minus_trailing
		ldmfd	SP!, {TMP0, TMP1}
		cmp	TOS, TMP0
		bls	1f
		mov	TOS, TMP0
		b	2f
1:		add	TMP1, TMP1, TOS
		sub	TOS, TMP0, TOS
2:		str	TMP1, [SP, #4]!
		NEXT
	CODE_END

	COLON_BEGIN	0, "BLANK", _blank, _slash_string
		.long	_lit, ' '
		.long	_fill
		.long	_exit
	COLON_END

	CODE_BEGIN	0, "CMOVE", _cmove, _blank
		ldmfd	SP!, {TMP0, TMP1}
		teq	TOS, #0
		beq	2f
1:		subs	TOS, TOS, #1
		ldrb	TMP2, [TMP0], #1
		strb	TMP2, [TMP1], #1
		bne	1b
2:		ldr	TOS, [SP], #4
		NEXT
	CODE_END
	
	CODE_BEGIN	0, "CMOVE>", _cmove_up, _cmove
		ldmfd	SP!, {TMP0, TMP1}
		teq	TOS, #0
		beq	2f
1:		subs	TOS, TOS, #1
		ldrb	TMP2, [TMP1, TOS]
		strb	TMP2, [TMP0, TOS]
		bne	1b
2:		ldr	TOS, [SP], #4
		NEXT
	CODE_END
	
#endif /* defined(STRING) */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Extra words (not a part of ANS or required by implementation)
 */

	COLON_BEGIN	0, "BINARY", _binary, _cmove_up
		.long	_lit, 2, _base, _store
		.long	_exit
	COLON_END

	/* (--) Use _code_gdb as a breakpoint for debugging */
	CODE_BEGIN	COMPILE_ONLY, "GDB", _gdb, _binary
		NEXT
	CODE_END

	/* Implementation specific - half word store */
	CODE_BEGIN	0, "H!", _h_store, _gdb
		ldmfd	SP!, {TMP0, TMP1}
		strh	TMP0, [TOS]
		mov	TOS, TMP1
		NEXT
	CODE_END

	/* Implementation specific - half word fetch */
	CODE_BEGIN	0, "H@", _h_fetch, _h_store
		ldrh	TOS, [TOS]
		NEXT
	CODE_END

	/* ( -- c-addr ) Push the address of RCS Id string */
	COLON_BEGIN	0, "RCSID", _rcsid, _h_fetch
		.long	_lit, 1f
		.long	_exit
		/* Not reached */
1:		CSTRING	"$Id: forth.S,v 1.171 2000/08/04 19:53:25 andrewm Exp $"
	COLON_END

	COLON_BEGIN	0, ".RCSID", _display_rcsid, _rcsid
		.long	_rcsid, _count, _type, _cr
		.long	_exit
	COLON_END

	COLON_BEGIN	0, "HELLO", _hello, _display_rcsid
		.long	_display_rcsid
		.long	_lit, 1f, _count, _cr, _type, _cr, _cr
		.long	_exit
		/* Not reached */
1:		CSTRING	"Welcome to StrongARM Forth!"
	COLON_END

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

	COLON_BEGIN	0, "QUIT", _quit, _hello
		/* Initialize RP */
		.long	_rp0, _fetch, _rp_store
		/* Initialize CP */
		.long	_align, _lit, CP_SIZE, _allot
		.long	_here, _dup, _cp_store, _cp0, _store
		/* Initialize TIB */
		.long	_here, _tib, _store
		.long	_lit, TIB_SIZE, _allot
		/* Enter interpretation state */
		.long	_lit, 0, _state, _store
		/* Splash screen */
		.long	_hello
		/* Read in a line */
1:		.long	_prompt
		.long	_tib, _fetch, _lit, TIB_SIZE, _accept
		.long	_number_tib, _store
		.long	_lit, 0, _from_in, _store

2:		.long	_bl, _word
		.long	_dup, _c_fetch
		.long	_question_branch, 6f
			.long	_find
			.long	_dup, _0_equal
			.long	_question_branch, 3f 	/* not found */
				.long	_drop
				.long	_lit, 0, _dup, _rot, _count
				.long	_to_number
				.long	_2drop, _drop
				.long	_state, _fetch
				.long	_question_branch, 2b
					.long	_literal
					.long	_branch, 2b
3:			
			.long	_0_less
			.long	_question_branch, 4f	/* immediate */
				.long	_execute
				.long	_branch, 2b
4:			.long	_state, _fetch		/* found */
			.long	_question_branch, 5f	/* compile */
				.long	_comma
				.long	_branch, 2b
5:
			.long	_execute
			.long	_branch, 2b
6:			.long	_drop
			.long	_branch, 1b
	COLON_END

/*
 *	: FIND ( c-addr -- c-addr 0 | xt 1 | xt -1 )
 *		>R
 *		LAST
 *		BEGIN
 *			@ DUP
 *		WHILE
 *			DUP CELL+ CHAR+ COUNT 2DUP
 *			R@ COUNT
 *			COMPARE 0=
 *			IF
 *				+ ALIGNED
 *				SWAP CELL+ C@
 *				LITERAL IMMEDIATE AND
 *				IF
 *					-1
 *					R> DROP
 *					EXIT
 *				ELSE
 *					1
 *					R> DROP
 *					EXIT
 *				THEN
 *			THEN
 *			2DROP
 *		REPEAT
 *		DROP R>
 *		0
 *	;
 */
	COLON_BEGIN	0, "FIND", _find, _quit
		.long	_to_r
		.long	_last
1:		.long	_fetch, _dup
		.long	_question_branch, 4f
			.long	_dup, _cell_plus, _char_plus, _count, _2dup
			.long	_r_fetch, _count
			.long	_compare, _0_equal
			.long	_question_branch, 3f
				.long	_plus, _aligned
				.long	_swap, _cell_plus, _c_fetch
				.long	_lit, IMMEDIATE, _and
				.long	_question_branch, 2f
					.long	_lit, -1
					.long	_r_from, _drop
					.long	_exit
2:				.long	_lit, 1
				.long	_r_from, _drop
				.long	_exit
3:			.long	_2drop
			.long	_branch, 1b
4:		.long	_drop, _r_from
		.long	_lit, 0
		.long	_exit
	COLON_END

	COLON_BEGIN	IMMEDIATE, "\073", _semicolon, _find
		.long	_lit, _exit, _comma
		.long	_c_from, _last, _fetch, _over, _store
		.long	_last, _store
		.long	_lit, S_INTERPRET, _state, _store
		.long	_exit
	COLON_END

	COLON_BEGIN	0, ":", _colon, _semicolon
		.long   _bl, _word, _c_fetch, _dup
		.long   _question_branch, 1f
			.long	_align
			.long	_here, _to_c		/* Save link pointer */
			.long	_lit, 0, _comma		/* Link */
			.long	_lit, 0, _c_comma	/* Flags */
			.long	_char_plus, _allot	/* Name cstring */
			.long	_here, _lit, 3, _lit, 0, _fill
			.long	_align
			.long	_lit, _code_enter, _comma
			.long	_lit, S_COMPILE, _state, _store
			.long	_exit
1:		.long	_drop
		.long   _lit, 0, _display_error
		/* Not reached */
        COLON_END

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

DICT_END
